// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{fs, path::Path, collections::HashMap};

use glium::{Surface, VertexBuffer, IndexBuffer, Display, Texture2d, Program, Frame, DrawParameters, index::PrimitiveType, uniforms::SamplerWrapFunction, uniform, implement_vertex};
use glutin::surface::WindowSurface;
use glam::UVec2;
use png::ColorType;

use super::{Blocks, mesh};

use crate::game::filesystem::Filesystem;
use crate::utils::texture;
use crate::playing::view::Camera;

#[derive(Clone, Copy)]
struct Vertex { pos: [f32; 2], tex_coords: [f32; 2] }
implement_vertex!(Vertex, pos, tex_coords);

pub struct Renderer {
	section_grid_size: UVec2,
	sections: HashMap<UVec2, Section>,
	ibo: IndexBuffer<u32>,
	texture: Texture2d,
	program: Program,
}

// Some if non-empty, None if empty
type Section = Option<VertexBuffer<Vertex>>;

pub(super) const SECTION_SIZE: u32 = 32;

impl Blocks {
	fn copy_section(&self, pos: UVec2) -> (UVec2, Box<[bool]>) {
		let min = pos * SECTION_SIZE;
		debug_assert!(min.cmplt(self.grid_size).all());

		let max = (min + UVec2::splat(SECTION_SIZE)).min(self.grid_size);

		let size = max - min;
		let mut section = Vec::with_capacity(size.element_product() as usize);

		let mut i = (min.x + min.y * self.grid_size.x) as usize;
		for _ in 0..size.y {
			section.extend_from_slice(&self.data[i..i + size.x as usize]);
			i += self.grid_size.x as usize;
		}

		(size, section.into_boxed_slice())
	}
}

impl Renderer {
	pub fn new(display: &Display<WindowSurface>, fs: &Filesystem) -> Renderer {
		let vsh = fs::read_to_string(fs.get(Path::new("shaders/blocks.vsh"))).unwrap();
		let fsh = fs::read_to_string(fs.get(Path::new("shaders/blocks.fsh"))).unwrap();

		let mut indices = Vec::with_capacity((SECTION_SIZE * SECTION_SIZE * 6) as usize);
		for i in (0..SECTION_SIZE * SECTION_SIZE * 4).step_by(4) {
			indices.extend_from_slice(&[
				i    , i + 1, i + 2,
				i + 1, i + 2, i + 3,
			]);
		}

		Renderer {
			section_grid_size: UVec2::ZERO,
			sections: HashMap::new(),
			ibo: IndexBuffer::immutable(display, PrimitiveType::TrianglesList, &indices).unwrap(),
			texture: texture::load(display, &fs.get(Path::new("textures/block.png")), ColorType::Rgb),
			program: Program::from_source(display, &vsh, &fsh, None).unwrap(),
		}
	}

	pub fn reset(&mut self, blocks: &Blocks) {
		self.section_grid_size = UVec2::new(blocks.grid_size.x.div_ceil(SECTION_SIZE), blocks.grid_size.y.div_ceil(SECTION_SIZE));
		self.sections.clear();
		self.sections.reserve(self.section_grid_size.element_product() as usize);
	}

	pub fn render(&mut self, display: &Display<WindowSurface>, frame: &mut Frame, params: &DrawParameters, blocks: &Blocks, camera: &Camera) {
		if let Some((min, max)) = self.get_visible_sections(blocks, camera) {
			let uniforms = uniform! { u_matrix: camera.get_matrix().to_cols_array_2d(), u_texture: self.texture.sampled().wrap_function(SamplerWrapFunction::Repeat) };

			for y in min.y..=max.y {
				for x in min.x..=max.x {
					let pos = UVec2::new(x, y);
					let section = self.sections.entry(pos).or_insert_with(|| Renderer::generate_section(pos, blocks, display));

					if let Some(section) = &*section {
						frame.draw(section, self.ibo.slice(0..(section.len() / 4) * 6).unwrap(), &self.program, &uniforms, params).unwrap();
					}
				}
			}
		}
	}

	fn generate_section(pos: UVec2, blocks: &Blocks, display: &Display<WindowSurface>) -> Section {
		let (size, section) = blocks.copy_section(pos);
		let num_blocks = section.iter().filter(|&&b| b).count();
		if num_blocks == 0 { return None; }

		let block_size = blocks.block_size.get() as f32;
		let off = (pos * SECTION_SIZE * blocks.block_size.get()).as_vec2() - blocks.grid_size.as_vec2() * block_size / 2.0;

		let rects = mesh::section_to_rects(size, section);
		let mut vertices = Vec::with_capacity(rects.len() * 4); // 4 vertices per rectangle (index buffer used)

		for rect in rects {
			let rect_size = (rect.1 - rect.0).as_vec2();
			let p0 = rect.0.as_vec2() * block_size + off;
			let p1 = p0 + rect_size * block_size;
			vertices.extend_from_slice(&[
				Vertex { pos: [p0.x, p0.y], tex_coords: [0.0, 0.0] },
				Vertex { pos: [p1.x, p0.y], tex_coords: [rect_size.x * block_size, 0.0] },
				Vertex { pos: [p0.x, p1.y], tex_coords: [0.0, rect_size.y * block_size] },
				Vertex { pos: [p1.x, p1.y], tex_coords: [rect_size.x * block_size, rect_size.y * block_size] },
			]);
		}

		Some(VertexBuffer::immutable(display, &vertices).unwrap())
	}

	fn get_visible_sections(&self, blocks: &Blocks, camera: &Camera) -> Option<(UVec2, UVec2)> {
		let block_size = blocks.block_size.get() as f32;
		let pos = camera.get_pos() + block_size * blocks.grid_size.as_vec2() / 2.0;
		let size = camera.get_size();

		let mut min = ((pos - size / 2.0) / (block_size * SECTION_SIZE as f32)).floor().as_ivec2();
		let mut max = ((pos + size / 2.0) / (block_size * SECTION_SIZE as f32)).floor().as_ivec2();

		for i in 0..2 {
			min[i] = min[i].max(0);
			max[i] = max[i].min(self.section_grid_size[i] as i32 - 1);

			if min[i] as u32 >= self.section_grid_size[i] || max[i] < 0 { return None; }
		}

		Some((min.as_uvec2(), max.as_uvec2()))
	}
}
