// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::collections::{BTreeMap, BTreeSet};

use serde::{Serialize, Deserialize};

#[cfg(feature = "client")] use crate::{world::MAX_PLAYERS, net::client::ClientError};

#[derive(Default, Clone, Serialize, Deserialize)]
pub struct Scores {
	scores: BTreeMap<u32, i32>,
	ordering: ScoreOrdering,
}

#[repr(u8)]
#[derive(Default, Clone, Copy, Serialize, Deserialize)]
#[serde(deny_unknown_fields)]
pub enum ScoreOrdering {
	Ascending,
	#[default] Descending,
}

impl Scores {
	pub fn new(ordering: ScoreOrdering) -> Scores {
		Scores { scores: BTreeMap::new(), ordering }
	}

	#[cfg(feature = "client")]
	pub fn validate(&self) -> Result<(), ClientError> {
		if self.scores.len() <= MAX_PLAYERS {
			Ok(())
		} else {
			Err(ClientError::Limit(String::from("too many scores")))
		}
	}

	#[cfg(feature = "client")]
	pub fn apply_delta(&mut self, delta: &DScores) {
		delta.apply(&mut self.scores);
	}

	pub fn get_scores(&self) -> &BTreeMap<u32, i32> { &self.scores }
	pub(super) fn get_scores_mut(&mut self) -> &mut BTreeMap<u32, i32> { &mut self.scores }
	#[cfg(feature = "client")] pub fn get_ordering(&self) -> ScoreOrdering { self.ordering }
}

#[derive(Clone, Serialize, Deserialize)]
pub struct DScores(BTreeMap<u32, Option<i32>>);

impl DScores {
	pub fn new(new: &BTreeMap<u32, i32>, old: &BTreeMap<u32, i32>) -> Option<DScores> {
		let mut delta = BTreeMap::new();

		// All scores removed
		let (new_ids, old_ids): (BTreeSet<u32>, BTreeSet<u32>) = (new.keys().copied().collect(), old.keys().copied().collect());

		let to_remove = old_ids.difference(&new_ids).copied();
		for id in to_remove {
			delta.insert(id, None);
		}

		// All scores added or changed
		for (&id, &score) in new {
			if old.get(&id).copied() != Some(score) {
				delta.insert(id, Some(score));
			}
		}

		(!delta.is_empty()).then_some(DScores(delta))
	}

	pub fn apply(&self, scores: &mut BTreeMap<u32, i32>) {
		for (&id, delta) in &self.0 {
			if let Some(score) = delta { scores.insert(id, *score); }
			else { scores.remove(&id); }
		}
	}
}

#[cfg(test)]
mod tests {
	use super::*;

	#[test]
	fn consistent() {
		for i in 0..10000 {
			println!("{i}");
			property(scores(), scores());
		}

		for _ in 0..100 {
			property(scores(), BTreeMap::new());
			property(BTreeMap::new(), scores());
		}

		property(BTreeMap::new(), BTreeMap::new());
	}

	fn scores() -> BTreeMap<u32, i32> {
		let mut scores = BTreeMap::new();
		for id in 0..16 {
			if rand::random::<bool>() {
				scores.insert(id, (rand::random::<u32>() % 16) as i32 - 8);
			}
		}
		scores
	}

	fn property(new: BTreeMap<u32, i32>, mut old: BTreeMap<u32, i32>) {
		if let Some(delta) = DScores::new(&new, &old) {
			delta.apply(&mut old);
		}
		assert_eq!(old, new);
	}
}
