// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::collections::BTreeMap;

#[derive(Default)]
pub struct Joined(BTreeMap<u32, JoinState>);

enum JoinState {
	NewClientPlaying, // A client has joined as a player
	OldClientPlaying, // A previously spectating client has started playing
	NewClientSpectating, // A client has joined as a spectator
}

impl Joined {
	pub(super) fn add_client(&mut self, id: u32, spectating: bool) {
		let prev = self.0.insert(id, if spectating { JoinState::NewClientSpectating } else { JoinState::NewClientPlaying });
		debug_assert!(prev.is_none());
	}

	pub(super) fn add_player(&mut self, id: u32) {
		self.0.entry(id).and_modify(|e| match e {
		  	// Don't want to forget that this is a new client, or the initial messages won't be sent out
			JoinState::NewClientPlaying | JoinState::OldClientPlaying => (),
			JoinState::NewClientSpectating => *e = JoinState::NewClientPlaying,
		}).or_insert(JoinState::OldClientPlaying);
	}

	pub(super) fn is_new_client(&self, id: u32) -> bool {
		match self.0.get(&id) {
			Some(JoinState::NewClientPlaying | JoinState::NewClientSpectating) => true,
			Some(JoinState::OldClientPlaying) | None => false,
		}
	}

	pub fn players(&self) -> impl Iterator<Item = u32> + '_ {
		self.0
			.iter()
			.filter(|(_id, state)| match state {
				JoinState::NewClientPlaying | JoinState::OldClientPlaying => true,
				JoinState::NewClientSpectating => false,
			})
			.map(|(&id, _state)| id)
	}
}
