// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::collections::VecDeque;

use crate::world::{MAX_STEPS, player::update::{PlayerUpdate, PlayerBulkUpdate}};
use crate::utils::RollingAverage;

/**
 * A queue storing player updates that can have "negative length".
 *
 * This queue can have "negative length" if it's popped from more times than it
 * is pushed to. If this happens, the previous player update that was pushed
 * would be returned.
 *
 * See `after_popped` for how this queue's length can grow and shrink to
 * minimise problems experienced by the client.
 */
pub(super) struct PlayerUpdateQueue {
	queue: VecDeque<(PlayerUpdate, u64)>,
	prev_update: (PlayerUpdate, u64),
	deficit: u64,
	received: u64, // How many updates received from any source (regular and redundant bulk updates)
	stream_received: u64, // How many updates received from PlayerUpdate messages on the stream
	increase_deficit: bool,
	past_remaining_updates: RollingAverage,
}

impl PlayerUpdateQueue {
	pub fn new() -> PlayerUpdateQueue {
		PlayerUpdateQueue {
			queue: VecDeque::new(),
			prev_update: (PlayerUpdate::default(), 0),
			deficit: 0,
			received: 0,
			stream_received: 0,
			increase_deficit: false,
			past_remaining_updates: RollingAverage::new(128.try_into().unwrap()),
		}
	}

	pub fn begin_iter(&mut self) {
		self.increase_deficit |= self.received > 0;
	}

	/**
	 * Makes adjustments to the queue based on the history of how many updates
	 * are remaining in the queue.
	 *
	 * The purpose of this is that whenever the queue is exhausted, the server
	 * needs to generate a player update that might not correspond to the
	 * intended update that would otherwise have been sent by the client.
	 *
	 * This can be a problem because this would result in there being slight
	 * jitter of the client's player which can disrupt their movement or might
	 * make them miss out on a shot.
	 *
	 * This problem is solved by selectively not increasing the deficit of
	 * updates when popping from the queue. This results in the queue's length
	 * settling on a value with some "wiggle room" to avoid jitter.
	 *
	 * However the longer the queue is, the greater the input lag which also is a
	 * problem. Therefore a trade-off needs to be made between jitter and input
	 * lag, with somewhat arbitrary parameters to get the best of both worlds.
	 */
	pub fn after_popped(&mut self) {
		const MIN: f32 = 1.0;
		const K_MIN: f32 = 3.5;
		const MAX: f32 = 4.0;
		const K_MAX: f32 = 8.0;

		// If too early to collect this data
		if !self.increase_deficit { return; }

		let amount_left = self.get_len();
		self.past_remaining_updates.add(amount_left as f32);
		let (mean, stddev) = self.past_remaining_updates.mean_and_stddev().unwrap();

		/*
		 * Avoids overestimating the standard deviation because the μ - kσ
		 * calculation is done using the assumption of a normal distribution, and
		 * it's possible that the client oscillates between neighbouring step
		 * counts, which shouldn't be considered.
		 */
		let stddev = (stddev - 0.5).max(0.0);

		let low_min = mean - K_MIN * stddev;
		let low_max = mean - K_MAX * stddev;

		if low_min < MIN {
			self.deficit = self.deficit.saturating_sub((MIN - low_min).ceil() as u64);
		} else if low_max > MAX {
			self.queue.drain(..((low_max - MAX).ceil() as usize).min(self.queue.len()));
		}
	}

	pub fn push(&mut self, update: PlayerUpdate, count: u64) {
		for _ in 0..count {
			self.stream_received += 1;
			if self.stream_received > self.received {
				self.update_received(update);
				assert_eq!(self.received, self.stream_received);
			}
		}
	}

	pub fn bulk_update(&mut self, bulk_update: PlayerBulkUpdate) {
		if let Some(bulk_update) = bulk_update.truncate(self.received + 1) {
			for update in bulk_update {
				self.update_received(update);
			}
		}
	}

	fn update_received(&mut self, update: PlayerUpdate) {
		self.received += 1;
		self.prev_update = (update, self.received);

		if self.deficit > 0 {
			self.deficit -= 1;
		} else if self.queue.len() >= MAX_STEPS {
			return;
		} else {
			self.queue.push_back(self.prev_update);
		}
	}

	pub fn pop(&mut self) -> (PlayerUpdate, u64) {
		self.queue.pop_front().unwrap_or_else(|| {
			if self.increase_deficit {
				self.deficit += 1;
			}
			self.prev_update
		})
	}

	pub fn get_len(&self) -> i64 {
		if self.queue.is_empty() {
			-(self.deficit as i64)
		} else {
			self.queue.len() as i64
		}
	}
}
