// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod controller;
pub(super) mod goal;

use std::collections::{BTreeSet, BTreeMap};

use glam::Vec2;

use controller::Controller;
use goal::Goal;

use super::config::ScoringConfig;

use crate::net::{scores::{Scores, DScores}, message::{Message, MiscMessage}};
use crate::world::{World, special_area::SpecialArea};

pub struct Scoring {
	prev_scores: BTreeMap<u32, i32>,
	scores: Scores,
	init_score: i32,
	controller: Controller,
	goal: Goal,
	elimination: Goal,
}

pub enum ScoringEvent {
	Kill(u32),
	Death(u32),
	Time(f32),
	Pos(u32, Vec2, f32),
}

pub struct ScoringUpdateResult {
	pub winners: Option<BTreeSet<u32>>,
	pub eliminated: Option<BTreeSet<u32>>,
}

impl Scoring {
	pub fn new(config: ScoringConfig, special_areas: &mut Vec<SpecialArea>) -> Scoring {
		Scoring {
			prev_scores: BTreeMap::default(),
			scores: Scores::new(config.ordering),
			init_score: config.init_score,
			controller: Controller::new(config.controller, config.min_score, config.max_score, special_areas),
			goal: config.goal,
			elimination: config.elimination,
		}
	}

	/**
	 * Clears all scores.
	 *
	 * Should be called whenever the game finishes and enters the lobby state.
	 */
	pub fn clear(&mut self) {
		self.scores.get_scores_mut().clear();
	}

	/**
	 * Resets all scores to their initial values.
	 *
	 * Should be called when the game starts.
	 */
	pub fn start(&mut self) {
		for score in self.scores.get_scores_mut().values_mut() {
			*score = self.controller.clamp_score(self.init_score);
		}

		self.goal.reset();
		self.controller.reset();
	}

	pub fn event(&mut self, event: ScoringEvent) {
		self.controller.event(event, self.scores.get_scores_mut());
	}

	pub fn update(&mut self, world: &World, time: Option<f64>) -> ScoringUpdateResult {
		// Removes old players and adds new players
		let scores = self.scores.get_scores_mut();
		scores.retain(|&id, _| world.has_player(id));

		for id in world.id_iter() {
			scores.entry(id).or_insert_with(|| self.controller.clamp_score(self.init_score));
		}

		let scores = self.scores.get_scores_mut();
		let eliminated = self.elimination.achieved(scores, time);

		// Avoids the final score flickering
		if let Some(eliminated) = &eliminated {
			for id in eliminated {
				scores.remove(id);
			}
		}

		/*
		 * Need to get the winners after the eliminated players in case the
		 * winning goal is `LastWith` and all players get eliminated at once. If I
		 * updated the winning goal first, the players would be removed with the
		 * game prematurely ending.
		 */
		ScoringUpdateResult { winners: self.goal.achieved(scores, time), eliminated }
	}

	pub fn init_scores_message(&self) -> Message {
		Message::Misc(MiscMessage::InitScores(self.scores.clone()))
	}

	pub fn delta_scores_message(&mut self) -> Option<Message> {
		DScores::new(self.scores.get_scores(), &self.prev_scores).map(|delta| {
			delta.apply(&mut self.prev_scores);
			Message::Misc(MiscMessage::DScores(delta))
		})
	}
}
