// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::mem;

use super::{ScoringEvent, BTreeMap};

use super::super::config::ScoringEventListener;

use crate::world::{player::RADIUS, special_area::SpecialArea};
use crate::utils::{Clock, maths::{RectCorners, CollidingRect}};

pub struct Controller {
	kill: i32,
	death: i32,
	clocks: Vec<(Clock, i32)>,
	occupation_zones: Vec<OccupationZone>,

	score_limits: ScoreLimits,
}

struct ScoreLimits {
	min: Option<i32>,
	max: Option<i32>,
}

impl ScoreLimits {
	pub fn clamp(&self, mut score: i32) -> i32 {
		if let Some(min) = self.min { score = score.max(min); }
		if let Some(max) = self.max { score = score.min(max); }
		score
	}

	fn change(&self, id: u32, dscore: i32, scores: &mut BTreeMap<u32, i32>) {
		if let Some(score) = scores.get_mut(&id) { // Should always be there but not using unwrap just in case
			*score = self.clamp(*score + dscore);
		}
	}
}

struct OccupationZone {
	owner: u32,
	rect: RectCorners,
	capture_time: f32,
	period: f32,
	dscore: i32,
	colliding: BTreeMap<u32, Clock>,
}

impl Controller {
	pub fn new(listeners: Vec<(ScoringEventListener, i32)>, min_score: Option<i32>, max_score: Option<i32>, special_areas: &mut Vec<SpecialArea>) -> Controller {
		let (mut kill, mut death, mut clocks, mut occupation_zones) = (0, 0, Vec::new(), Vec::new());

		for (listener, dscore) in listeners {
			match listener {
				ScoringEventListener::Kill => kill += dscore,
				ScoringEventListener::Death => death += dscore,
				ScoringEventListener::Time(time) => clocks.push((Clock::new(time), dscore)),
				ScoringEventListener::Occupation(mut zone, colour) => {
					for i in 0..2 {
						if zone.p0[i] > zone.p1[i] {
							mem::swap(&mut zone.p0[i], &mut zone.p1[i]);
						}
					}

					if let Some(colour) = colour {
						special_areas.push(SpecialArea { pos: zone.p0, size: zone.p1 - zone.p0, colour: colour.get(zone.owner) });
					}

					occupation_zones.push(OccupationZone {
						owner: zone.owner,
						rect: RectCorners(zone.p0, zone.p1),
						capture_time: zone.capture_time,
						period: zone.period,
						dscore,
						colliding: BTreeMap::new(),
					});
				},
			}
		}

		Controller { kill, death, clocks, occupation_zones, score_limits: ScoreLimits { min: min_score, max: max_score } }
	}

	pub fn reset(&mut self) {
		for (clock, _) in &mut self.clocks {
			clock.set_next_update(clock.get_interval());
		}

		for zone in &mut self.occupation_zones {
			zone.colliding.clear();
		}
	}

	pub fn clamp_score(&self, score: i32) -> i32 {
		self.score_limits.clamp(score)
	}

	pub fn event(&mut self, event: ScoringEvent, scores: &mut BTreeMap<u32, i32>) {
		match event {
			ScoringEvent::Kill(id) => self.score_limits.change(id, self.kill, scores),
			ScoringEvent::Death(id) => self.score_limits.change(id, self.death, scores),
			ScoringEvent::Time(dt) => {
				let mut dscore = 0;
				for (clock, ds) in &mut self.clocks {
					let steps = clock.advance(dt);
					dscore += *ds * steps as i32;
				}

				for score in scores.values_mut() {
					*score = self.score_limits.clamp(*score + dscore);
				}
			},
			ScoringEvent::Pos(id, pos, dt) => {
				for zone in &mut self.occupation_zones {
					// Can't capture your own occupation zone
					if id == zone.owner { continue; }

					let colliding = (pos, RADIUS).colliding_rect(&zone.rect);

					if colliding {
						let clock = zone.colliding.entry(id).or_insert_with(|| Clock::with_next_update(zone.period, zone.capture_time));
						let steps = clock.advance(dt);
						if steps > 0 {
							self.score_limits.change(id, steps as i32 * zone.dscore, scores);
						}
					} else {
						zone.colliding.remove(&id);
					}
				}
			},
		}
	}
}
