// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{cmp::Ordering, num::NonZeroUsize};

use serde::Deserialize;

use super::{BTreeSet, BTreeMap};

#[derive(Default, Deserialize)]
#[serde(deny_unknown_fields)]
pub enum Goal {
	/**
	 * Finds all players whose scores satisfy the given condition, returning
	 * Some(players) if players isn't empty. Otherwise returns None.
	 */
	FirstTo(Condition<i32>),

	/**
	 * Finds all player whose scores satisfy the given condition.
	 *
	 * If there are multiple players with that condition, return None. If there
	 * is just one player with that condition, return Some(player). If there are
	 * no players with that condition, return Some(prev_players), where
	 * `prev_players` is the previous set of players satisfying this condition
	 * (this set is initialised to be empty).
	 */
	LastWith(Condition<i32>, #[serde(skip)] BTreeSet<u32>),

	/**
	 * Finds the players with the maximum or minimum score.
	 *
	 * If there are no players in the game, returns None.
	 *
	 * If `exclusive` is set to true and there are multiple players with the same
	 * extreme score, returns None.
	 */
	Extremum {
		#[serde(rename = "type")] typ: ExtremumType,
		exclusive: bool,
	},

	/**
	 * Returns Some(all players) once the condition with the timer's time (in
	 * seconds) has been met, otherwise returning None.
	 *
	 * Note that if the timer has a limit to prevent it being displayed above or
	 * below a certain value, the time clamped with that limit will be used.
	 *
	 * If the timer is None, this will always return None.
	 */
	Timer(Condition<f64>),

	/**
	 * Combines multiple goals into a single goal.
	 *
	 * This first calls `achieved` on all child goals and if the number of
	 * achieved goals is greater than or equal to the necessary count, these sets
	 * of achieved players are reduced (through union or intersection) and
	 * Some(reduced set) is returned. Otherwise returns None if not enough goals
	 * are met.
	 *
	 * See the `And` and `Or` variants for special cases of this.
	 */
	Multiple {
		goals: Vec<Goal>,
		necessary: NonZeroUsize,
		reduction: SetReduction,
	},

	/**
	 * Returns a set of players if all goals are achieved.
	 *
	 * The intersection of all these sets is returned.
	 *
	 * This is identical to Multiple { goals, necessary:
	 * NonZeroUsize::new(goals.len())?, reduction: SetReduction::Intersection }.
	 *
	 * Always None if there are no goals.
	 */
	And(Vec<Goal>),

	/**
	 * Returns a set of players if at least one goal is achieved.
	 *
	 * The union of all these sets is returned.
	 *
	 * This is identical to Multiple { goals, necessary: 1, reduction:
	 * SetReduction::Union }.
	 *
	 * Always None if there are no goals.
	 */
	Or(Vec<Goal>),

	/**
	 * Always returns Some(all players).
	 */
	SomeAll,

	/**
	 * Always returns Some(empty set).
	 */
	SomeEmpty,

	/**
	 * The game never ends, always returns None.
	 */
	#[default] None,
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub struct Condition<T>(Comparison, T);

#[derive(PartialEq, Eq, Deserialize)]
pub enum Comparison {
	Less,
	LessEqual,
	Equal,
	NotEqual,
	GreaterEqual,
	Greater,
}

impl<T> Condition<T> where T: PartialOrd {
	fn test(&self, lhs: T) -> bool {
		let Some(cmp) = lhs.partial_cmp(&self.1) else {
			// Can happen if there's a NaN, in which op(NaN, x) returns false unless op is `!=`
			return self.0 == Comparison::NotEqual;
		};

		match self.0 {
			Comparison::Less => cmp == Ordering::Less,
			Comparison::LessEqual => cmp != Ordering::Greater,
			Comparison::Equal => cmp == Ordering::Equal,
			Comparison::NotEqual => cmp != Ordering::Equal,
			Comparison::GreaterEqual => cmp != Ordering::Less,
			Comparison::Greater => cmp == Ordering::Greater,
		}
	}
}

impl Condition<i32> {
	fn players_with(&self, scores: &BTreeMap<u32, i32>) -> BTreeSet<u32> {
		scores.iter().filter_map(|(&id, &score)| self.test(score).then_some(id)).collect()
	}
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub enum ExtremumType {
	Minimum,
	Maximum,
}

#[derive(Clone, Copy, Deserialize)]
#[serde(deny_unknown_fields)]
pub enum SetReduction {
	Union,
	Intersection,
}

impl Goal {
	/**
	 * Clears the state of the `LastWith` goal.
	 */
	pub fn reset(&mut self) {
		match self {
			Goal::LastWith(_, prev_with_cond) => prev_with_cond.clear(),
			Goal::Multiple { goals, .. } => goals.iter_mut().for_each(Goal::reset),
			_ => (),
		}
	}

	/**
	 * Returns a set of players who have achieved the goal, or None.
	 *
	 * This is used in the context of both the win condition and the condition
	 * for eliminating players. For the win condition, Some(empty set) means that
	 * the game ends with no winners while None means the game is still going.
	 * For eliminating players, Some(empty set) and None are the same.
	 */
	pub fn achieved(&mut self, scores: &BTreeMap<u32, i32>, time: Option<f64>) -> Option<BTreeSet<u32>> {
		match self {
			Goal::FirstTo(cond) => {
				let achieved = cond.players_with(scores);
				(!achieved.is_empty()).then_some(achieved)
			},
			Goal::LastWith(cond, prev_with_cond) => {
				let with_cond = cond.players_with(scores);

				let ret = match with_cond.len() {
					1 => Some(with_cond.clone()),
					0 => Some(prev_with_cond.clone()),
					_ => None,
				};

				if !with_cond.is_empty() {
					*prev_with_cond = with_cond;
				}

				ret
			}
			Goal::Extremum { typ, exclusive } => {
				let Some(extremum) = match typ {
					ExtremumType::Minimum => scores.values().min(),
					ExtremumType::Maximum => scores.values().max(),
				}.copied() else {
					return None; // If no players
				};

				let achieved: BTreeSet<u32> = scores.iter().filter_map(|(&id, &score)| (score == extremum).then_some(id)).collect();
				debug_assert!(!achieved.is_empty());

				if *exclusive && achieved.len() > 1 {
					return None;
				}

				Some(achieved)
			},
			Goal::Timer(cond) => time.and_then(|time| cond.test(time).then(|| scores.keys().copied().collect())),
			Goal::Multiple { goals, necessary, reduction } => Goal::multiple_goals(scores, time, goals, *necessary, *reduction),
			Goal::And(goals) => {
				let len = NonZeroUsize::new(goals.len())?;
				Goal::multiple_goals(scores, time, goals, len, SetReduction::Intersection)
			},
			Goal::Or(goals) => Goal::multiple_goals(scores, time, goals, const { NonZeroUsize::new(1).unwrap() }, SetReduction::Union),
			Goal::SomeAll => Some(scores.keys().copied().collect()),
			Goal::SomeEmpty => Some(BTreeSet::new()),
			Goal::None => None,
		}
	}

	fn multiple_goals(scores: &BTreeMap<u32, i32>, time: Option<f64>, goals: &mut [Goal], necessary: NonZeroUsize, reduction: SetReduction) -> Option<BTreeSet<u32>> {
		let achieved_sets = goals.iter_mut().filter_map(|goal| goal.achieved(scores, time)).collect::<Vec<_>>();

		match reduction {
			_ if achieved_sets.len() < necessary.get() => None,
			SetReduction::Union => achieved_sets.into_iter().reduce(|a, b| a.union(&b).copied().collect()),
			SetReduction::Intersection => achieved_sets.into_iter().reduce(|a, b| a.intersection(&b).copied().collect()),
		}
	}
}
