// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{fmt::Write, collections::BTreeSet};

use strum::IntoEnumIterator;

use super::{Server, Client, SpectatingMode, DMap, SANITY_CHECK_PERIOD, CHAT_MAX_BURST, CHAT_RATE_LIMIT, MoveChooser, scoring::ScoringEvent, command::{Command, CommandVariant}, joined::Joined};

use crate::net::message::{Message, WorldMessage, MiscMessage};
use crate::net::message_stream::MsError;
use crate::net::text_io::{self, TextIoString, TextIoStr, ClientTextInput, ServerTextOutput};
use crate::world::{DELTA_TIME, update::WorldUpdate, changes::ServerChanges, event::WorldEvent};
use crate::utils::{smallest_unused, maths::Circle, command::{Command as _, CommandVariant as _}};

impl Server {
	/**
	 * Returns Ok(()) if everything's fine, and Err(()) if the server should
	 * close.
	 */
	pub fn update(&mut self, mut dt: f32, mut messages: Vec<Message>) -> Result<(), ()> {
		/*
		 * Need at least one step to prevent a player joining being added to
		 * `joined` but no world updates occurring for them to be added to the
		 * world. Then the `InitWorld` message is sent to the client without
		 * containing the player, which is a violation of the protocol and so the
		 * client leaves the game.
		 */
		let steps = self.clock.advance(dt);
		if steps == 0 { return Ok(()); }

		/*
		 * Updates delta time to prevent time running slower (for some delta times
		 * being completely ignored) for chat rate limiting and sending out sanity
		 * checks.
		 */
		dt = steps as f32 * DELTA_TIME;

		let prev_spectator_count = self.spectator_count();

		let mut joined = self.accept_new_clients()?;
		self.receive_client_messages(&mut messages, &mut joined, dt);
		self.remove_clients();

		for i in 0..steps {
			self.update_world(&mut messages, &joined, i == 0);

			if let Some(timer) = &mut self.timer {
				timer.update(DELTA_TIME);
			}
		}

		self.after_update(&messages, &joined, prev_spectator_count, self.spectator_count());
		if let Some(msg) = self.get_sanity_check(dt) {
			messages.push(msg);
		}

		if self.sync_timer {
			messages.push(Message::World(WorldMessage::Timer(self.timer.clone())));
			self.sync_timer = false;
		}

		self.send_messages(messages);
		Ok(())
	}

	fn spectator_count(&self) -> usize {
		self.clients.values().filter(|client| client.spectating.is_some()).count()
	}

	fn accept_new_clients(&mut self) -> Result<Joined, ()> {
		let mut joined = Joined::default();

		let mut player_count = self.world.player_count();
		loop {
			match self.connection_listener.get() {
				Ok(Some(incoming)) => {
					if player_count < self.max_players {
						let spectate =
							if incoming.spectate_requested() { Some(SpectatingMode::Manual) }
							else if self.can_join() { None }
							else { Some(SpectatingMode::Automatic) };

						let stream = match incoming.accept() {
							Ok(stream) => stream,
							Err((err, source)) => {
								log::info!("failed accepting player: {err}, source = {source}");
								continue;
							},
						};

						let id = smallest_unused::find(&self.clients);
						let source = stream.source();
						self.clients.insert(id, Client::new(stream, spectate));
						joined.add_client(id, spectate.is_some());
						log::info!("player (id = {id}) joined, spectating = {spectate:?}, source = {source}");

						player_count += 1;
					} else {
						incoming.reject(format!("too many players, limit is {}", self.max_players));
					}
				},
				Ok(None) => break,
				Err(()) => return Err(()),
			}
		}

		Ok(joined)
	}

	fn receive_client_messages(&mut self, messages: &mut Vec<Message>, joined: &mut Joined, dt: f32) {
		let can_join = self.can_join();

		for (&id, client) in self.clients.iter_mut().filter(|(_, client)| !client.to_remove) {
			client.updates.begin_iter();
			client.allowed_chats = (client.allowed_chats + CHAT_RATE_LIMIT * dt).min(CHAT_MAX_BURST);

			client.to_remove |= loop {
				match client.stream.receive() {
					Ok(Some(Message::PlayerUpdate(update, steps))) => client.updates.push(update, steps),
					Ok(Some(Message::PlayerBulkUpdate(bulk_update))) => client.updates.bulk_update(bulk_update),
					Ok(Some(Message::Misc(MiscMessage::ClientTextInput(ClientTextInput::Chat(msg))))) => {
						if client.allowed_chats >= 1.0 {
							log::info!("player (id = {id}) sent chat message: {}", msg.get());
							messages.push(Message::Misc(MiscMessage::ServerTextOutput(ServerTextOutput::Chat(id, msg))));
							client.allowed_chats -= 1.0;
						} else {
							client.per_player_messages.push((messages.len(), Message::Misc(MiscMessage::ServerTextOutput(ServerTextOutput::Server(text_io::from_const!("Rate limit reached, slow down."))))));
						}
					},
					Ok(Some(Message::Misc(MiscMessage::ClientTextInput(ClientTextInput::Command(cmd))))) => {
						let response = match Command::build::<CommandVariant>(&cmd) {
							Ok(Some(Command::Help(None))) => Some(TextIoString::new_truncate(CommandVariant::iter().map(CommandVariant::help).collect::<Vec<_>>().join("\n"))),
							Ok(Some(Command::Help(Some(cmd)))) => Some(TextIoString::new_truncate(String::from(cmd.help()))),
							Ok(Some(Command::Spectate)) => {
								let msg = Some(match client.spectating {
									Some(SpectatingMode::Manual) => text_io::from_const!("Already spectating."),
									Some(SpectatingMode::Automatic) => text_io::from_const!("Already spectating, and will continue to spectate when the game ends."),
									None => text_io::from_const!("Now spectating."),
								});
								client.spectating = Some(SpectatingMode::Manual);
								msg
							},
							Ok(Some(Command::Play)) => {
								Some(if let Some(mode) = client.spectating {
									match mode {
										_ if can_join => {
											client.spectating = None;
											joined.add_player(id);
											text_io::from_const!("Now playing.")
										},
										SpectatingMode::Manual => {
											client.spectating = Some(SpectatingMode::Automatic);
											text_io::from_const!("Will play once the game ends (joining mid-game has been disabled).")
										},
										SpectatingMode::Automatic => text_io::from_const!("Already chose to leave spectator mode once the game ends (joining mid-game has been disabled)."),
									}
								} else {
									text_io::from_const!("Already playing.")
								})
							},
							Ok(Some(Command::Ready)) => {
								if self.lobby.active() {
									if client.spectating.is_some() {
										Some(text_io::from_const!("Cannot vote ready while spectating."))
									} else {
										let success = self.lobby.add_ready(id);
										if success {
											None
										} else {
											Some(text_io::from_const!("Already voted."))
										}
									}
								} else {
									Some(text_io::from_const!("Game has already started."))
								}
							},
							Ok(Some(Command::Underpower(cmd))) => Some(self.event_chooser.underpower_command(id, cmd)),
							Ok(None) => None,
							Err(err) => Some(TextIoString::new_truncate(format!("error: {err}"))),
						};

						let before = messages.len();
						client.per_player_messages.push((before, Message::Misc(MiscMessage::ServerTextOutput(ServerTextOutput::Command(cmd)))));
						if let Some(response) = response {
							client.per_player_messages.push((before, Message::Misc(MiscMessage::ServerTextOutput(ServerTextOutput::CommandResponse(response)))));
						}
					},
					Ok(Some(_)) => { // Violation of the protocol so removed
						log::warn!("player (id = {id}) violated the protocol, sending an invalid message");
						break true;
					}
					Ok(None) => break false,
					Err(err) => {
						if let MsError::Other(err) = err {
							log::warn!("player (id = {id}) disconnected with an error: {err}");
						}
						break true;
					}
				}
			}
		}
	}

	fn remove_clients(&mut self) {
		self.clients.retain(|&id, client| {
			if client.to_remove {
				log::info!("player (id = {id}) left");
			}
			!client.to_remove
		});
	}

	fn update_world(&mut self, messages: &mut Vec<Message>, joined: &Joined, first: bool) {
		// Creating the player update map and all acknowledgements
		let (player_updates, acks) = {
			let mut updates = DMap::new();
			let mut acks = DMap::new();

			for (&id, client) in &mut self.clients {
				let (update, ack) = client.updates.pop();
				if client.spectating.is_none() {
					updates.insert(id, update);
				}
				acks.insert(id, ack);
			}

			(updates, acks)
		};

		// Constructing the pre_events
		let mut changes = ServerChanges::default();
		let move_chooser: &dyn MoveChooser = if self.lobby.active() { self.lobby.get_move_chooser() } else { &self.playing_move_chooser };
		let pre_events = self.event_chooser.get_pre_events(&self.world, &self.blocks, move_chooser, first.then_some(joined));
		self.world.apply_events(&pre_events, &mut changes);

		// Updating the world
		self.world.update(&player_updates, &self.blocks, &mut changes);

		// Updating the scoring
		self.scoring.event(ScoringEvent::Time(DELTA_TIME));
		for kill in &changes.kills {
			if let Some(killer) = kill.killer {
				self.scoring.event(ScoringEvent::Kill(killer));
			}
			self.scoring.event(ScoringEvent::Death(kill.killed));
		}

		for (&id, player) in self.world.get_players() {
			self.scoring.event(ScoringEvent::Pos(id, player.get_pos(), DELTA_TIME));
		}

		// Constructing the post_events
		let (mut post_events, spectator_changes) = self.get_post_events(messages, &changes);

		// Adds and removes players if necessary
		let mut player_change_events = Vec::new();
		for (id, new_spectating) in spectator_changes {
			let Some(client) = self.clients.get_mut(&id) else { continue; };

			if client.spectating.is_some() != new_spectating.is_some() {
				if new_spectating.is_some() {
					player_change_events.push(WorldEvent::RemovePlayer(id));
				} else {
					let move_chooser: &dyn MoveChooser = if self.lobby.active() { self.lobby.get_move_chooser() } else { &self.playing_move_chooser };
					self.event_chooser.add_player(id, &self.world, &self.blocks, move_chooser, &mut player_change_events);
				}
			}

			client.spectating = new_spectating;
		}
		self.world.apply_events(&player_change_events, &mut changes);
		post_events.append(&mut player_change_events);

		// Constructing the message for this update
		for (id, ack) in acks {
			if !joined.is_new_client(id) {
				let update = Message::World(WorldMessage::WorldUpdate(WorldUpdate {
					pre_events: pre_events.clone(), players: player_updates.clone(), post_events: post_events.clone(),
				}, ack));

				self.clients.get_mut(&id).unwrap().per_player_messages.push((messages.len(), update));
			}
		}
	}

	fn get_post_events(&mut self, messages: &mut Vec<Message>, changes: &ServerChanges) -> (Vec<WorldEvent>, DMap<u32, Option<SpectatingMode>>) {
		let mut spectator_changes = DMap::new();

		if self.lobby.active() {
			let just_started = self.lobby.just_started();
			for msg in self.lobby.update(DELTA_TIME, &self.world) {
				messages.push(Message::Misc(MiscMessage::ServerTextOutput(ServerTextOutput::Server(msg))));
			}

			// Resends the timer as after updating it is no longer paused
			if just_started {
				let timer = self.lobby.get_timer();
				self.set_timer(timer);
			}

			if self.world.player_count() < self.lobby.get_min_players() {
				self.reset_lobby(true);
			} else if !self.lobby.active() {
				self.transition_playing();
				return (self.reset(), spectator_changes);
			}
		} else {
			let res = self.scoring.update(&self.world, self.timer.as_ref().map(|timer| timer.displayed_seconds(timer.raw_seconds(), 0.0)));
			if let Some(eliminated) = res.eliminated {
				for id in eliminated {
					spectator_changes.insert(id, Some(SpectatingMode::Automatic));
				}
			}

			if let Some(winners) = res.winners {
				self.get_win_messages(winners, messages);
				self.transition_lobby(&mut spectator_changes);
				return (self.reset(), spectator_changes);
			} else if self.world.player_count() == 0 && self.lobby.reset_if_no_players() {
				self.transition_lobby(&mut spectator_changes);
				return (self.reset(), spectator_changes);
			}
		}

		let move_chooser: &dyn MoveChooser = if self.lobby.active() { self.lobby.get_move_chooser() } else { &self.playing_move_chooser };
		(self.event_chooser.get_and_apply_post_events(&mut self.world, &self.blocks, move_chooser, changes), spectator_changes)
	}

	fn transition_playing(&mut self) {
		self.scoring.start();
		self.set_timer(self.playing_timer.clone());
	}

	fn reset_lobby(&mut self, keep_ready_players: bool) {
		self.lobby.reset(&self.blocks, keep_ready_players);
		self.set_timer(self.lobby.get_timer());
	}

	fn transition_lobby(&mut self, spectator_changes: &mut DMap<u32, Option<SpectatingMode>>) {
		self.reset_lobby(false);

		for (&id, client) in &self.clients {
			if client.spectating != Some(SpectatingMode::Manual) {
				spectator_changes.insert(id, None);
			}
		}

		if self.lobby.active() {
			self.scoring.clear();
		} else { // If the lobby is disabled, instead go to the playing state
			self.transition_playing();
		}
	}

	fn get_win_messages(&mut self, winners: BTreeSet<u32>, messages: &[Message]) {
		let msg = match winners.len() {
			0 => String::from("Nobody won the game!"),
			1 => format!("Player {} won the game!", winners.first().unwrap() + 1),
			_ => {
				let mut msg = String::from("Players ");
				for (i, &id) in winners.iter().enumerate() {
					if i == winners.len() - 1 {
						let _ = write!(msg, " and ");
					} else if i > 0 {
						let _ = write!(msg, ", ");
					}

					let _ = write!(msg, "{}", id + 1);
				}
				let _ = write!(msg, " won the game!");
				msg
			},
		};

		for (id, client) in &mut self.clients {
			let msg = if winners.len() >= self.world.player_count() {
				String::from(
					if client.spectating.is_some() { "Everyone won!" }
					else { "Congratulations! Everyone won!" }
				)
			} else if winners.contains(id) {
				let mut other_winners = winners.clone();
				other_winners.remove(id);

				match other_winners.len() {
					0 => String::from("Congratulations! You won!"),
					1 => format!("Congratulations! You and player {} won!", other_winners.first().unwrap() + 1),
					_ => {
						let mut msg = String::from("Congratulations! You and players ");
						for (i, &id) in other_winners.iter().enumerate() {
							if i == other_winners.len() - 1 {
								let _ = write!(msg, " and ");
							} else if i > 0 {
								let _ = write!(msg, ", ");
							}

							let _ = write!(msg, "{}", id + 1);
						}
						let _ = write!(msg, " won!");
						msg
					},
				}
			} else {
				msg.clone()
			};

			client.per_player_messages.push((messages.len(), Message::Misc(MiscMessage::ServerTextOutput(ServerTextOutput::Server(TextIoString::new_truncate(msg))))));
		}
	}

	fn after_update(&mut self, messages: &[Message], joined: &Joined, prev_spectator_count: usize, spectator_count: usize) {
		let dscore_msg = self.scoring.delta_scores_message();

		for (&id, client) in &mut self.clients {
			// Balances jitter and input lag in the player queues
			client.updates.after_popped();

			let before = messages.len();
			if joined.is_new_client(id) {
				// Adds the initial messages if the player just joined
				client.per_player_messages.extend([
					Message::World(WorldMessage::InitWorld {
						player_id: id,
						world: self.world.clone(),
						blocks: self.compressed_blocks.clone(),
					}),
					Message::World(WorldMessage::Timer(self.timer.clone())),
					self.scoring.init_scores_message(),
				].map(|x| (before, x)));

				if let Some(msg) = &self.instructions {
					client.per_player_messages.push((before, Message::Misc(MiscMessage::ServerTextOutput(ServerTextOutput::Server(msg.clone())))));
				}
			} else if let Some(msg) = &dscore_msg {
				client.per_player_messages.push((before, msg.clone()));
			}

			if joined.is_new_client(id) || prev_spectator_count != spectator_count {
				client.per_player_messages.push((messages.len(), Message::Misc(MiscMessage::SpectatorCount(spectator_count))));
			}
		}
	}

	fn get_sanity_check(&mut self, dt: f32) -> Option<Message> {
		self.next_sanity_check -= dt;
		(self.next_sanity_check <= 0.0).then(|| {
			self.next_sanity_check += SANITY_CHECK_PERIOD;
			Message::World(WorldMessage::SanityCheck(self.world.clone()))
		})
	}

	fn send_messages(&mut self, messages: Vec<Message>) {
		for (&id, client) in self.clients.iter_mut().filter(|(_, client)| !client.to_remove) {
			if let Err(err) = client.send_messages(&messages) {
				client.to_remove = true;
				if let MsError::Other(err) = err {
					log::warn!("player (id = {id}) disconnected with an error: {err}");
				}
			}
		}
	}
}
