// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::ops::Deref;

use serde::{Serialize, Deserialize};

#[derive(Clone, Serialize, Deserialize)]
pub enum ClientTextInput {
	Chat(TextIoString),
	Command(TextIoString),
}

#[derive(Clone, Serialize, Deserialize)]
pub enum ServerTextOutput {
	Chat(u32, TextIoString), // Communication between players
	Admin(TextIoString), // A human-generated message from the server admin
	Server(TextIoString), // An automatic message from the server
	Command(TextIoString), // The exact command a user entered, the server sends this to acknowledge receiving a command
	CommandResponse(TextIoString), // An automatic message from the server responding to a command
}

pub const MAX_LEN: usize = 4096; // Within the limit the message streams accept

// INVARIANT: self.0.len() <= MAX_LEN
pub struct TextIoStr<'a>(&'a str);

impl TextIoStr<'_> {
	pub const fn new(s: &str) -> Option<TextIoStr> {
		if s.len() <= MAX_LEN { Some(TextIoStr(s)) }
		else { None }
	}

	pub fn owned(&self) -> TextIoString {
		TextIoString(String::from(self.0))
	}
}

// INVARIANT: self.0.len() <= MAX_LEN
#[derive(Default, Clone, Serialize, Deserialize)]
#[serde(try_from = "String", into = "String")]
pub struct TextIoString(String);

macro_rules! from_const {
	( $s:expr ) => {
		{ const { TextIoStr::new($s).unwrap() }.owned() }
	};
}

pub(crate) use from_const;

impl TextIoString {
	pub fn new() -> TextIoString {
		TextIoString(String::new())
	}

	pub fn new_truncate(mut s: String) -> TextIoString {
		const ELLIPSIS: char = '…';
		const ELLIPSIS_LEN: usize = ELLIPSIS.len_utf8();

		if s.len() <= MAX_LEN {
			return TextIoString(s);
		}

		while s.len() > MAX_LEN - ELLIPSIS_LEN {
			s.pop();
		}

		s.push(ELLIPSIS);
		TextIoString(s)
	}

	pub fn get(&self) -> &str {
		&self.0
	}

	pub fn try_push(&mut self, ch: char) {
		if self.0.len() + ch.len_utf8() <= MAX_LEN {
			self.0.push(ch);
		}
	}

	pub fn pop(&mut self) {
		self.0.pop();
	}
}

impl TryFrom<String> for TextIoString {
	type Error = &'static str;

	fn try_from(s: String) -> Result<TextIoString, &'static str> {
		(s.len() <= MAX_LEN).then_some(TextIoString(s)).ok_or("too long")
	}
}

impl From<TextIoString> for String {
	fn from(s: TextIoString) -> String {
		s.0
	}
}

impl Deref for TextIoString {
	type Target = str;

	fn deref(&self) -> &str {
		self.0.as_ref()
	}
}
