// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod renderer;

pub use renderer::Renderer as HealthBarRenderer;

use glam::{Vec2, Vec3};

use renderer::Instance;

use super::Camera;

use crate::world::{player::Player, smooth::LocalPlayer};
use crate::utils::maths::decay::Decay;

pub struct HealthBar {
	rendered_health: f32,
}

const HUD_WIDTH: f32 = 200.0;
pub const HUD_HEIGHT: f32 = 15.0;

const ABOVE_WIDTH: f32 = 1.0;
const ABOVE_HEIGHT: f32 = 0.125;

pub struct AbovePlayerBuilder<'a> {
	lwinsize: Vec2,
	size: [f32; 2],
	camera: &'a Camera,
}

impl HealthBar {
	pub fn new(health: f32) -> HealthBar { HealthBar { rendered_health: health.clamp(0.0, 1.0) } }
	pub fn reset(&mut self) { self.rendered_health = 1.0; }

	pub fn update(&mut self, player: &Player, dt: f32) {
		let health = player.get_norm_health();

		/*
		 * Decays the health bar faster when the health's lower as up-to-date
		 * information is more important then, which can prevent the player from
		 * overestimating their or others' health.
		 */
		let rate = 1.0 / health.max(0.0625) + 7.0;
		self.rendered_health.decay_to(health, rate, dt);
		self.rendered_health = self.rendered_health.clamp(0.0, 2.0);
	}

	/**
	 * Returns the x-pixel coordinates of the positionh and size of the health
	 * bar rendered on the HUD.
	 */
	pub fn hud_x_px(width: f32) -> (f32, f32) {
		let size = (width - 2.0 * HUD_HEIGHT).clamp(0.0, HUD_WIDTH);
		let pos = (width - size) / 2.0;
		(pos, size)
	}

	pub fn make_hud(&self, lwinsize: Vec2) -> Instance {
		let (px_x, px_width) = HealthBar::hud_x_px(lwinsize.x);
		let ndc_width = 2.0 * px_width / lwinsize.x;
		let ndc_x = 2.0 * px_x / lwinsize.x - 1.0;

		Instance {
			i_pos: [ndc_x, -1.0 + 2.0 * HUD_HEIGHT / lwinsize.y],
			i_size: [ndc_width, 2.0 * HUD_HEIGHT / lwinsize.y],
			i_t: self.get_t(),
			i_colour: self.get_colour(),
		}
	}

	pub fn above_player_builder(lwinsize: Vec2, camera: &Camera) -> AbovePlayerBuilder {
		let size = camera.get_size();

		AbovePlayerBuilder {
			lwinsize, camera,
			size: [
				2.0 * ABOVE_WIDTH / size.x,
				2.0 * ABOVE_HEIGHT / size.y,
			],
		}
	}

	fn get_colour(&self) -> [f32; 3] {
		let t = self.rendered_health;
		if t < 0.5 {
			Vec3::new(1.0, t * 2.0, 0.0)
		} else if t <= 1.0 {
			Vec3::new(2.0 - t * 2.0, 1.0, 0.0)
		} else {
			Vec3::new(0.0, 0.75, 1.0)
		}.to_array()
	}

	fn get_t(&self) -> f32 {
		if self.rendered_health <= 1.0 { self.rendered_health } else { self.rendered_health - 1.0 }
	}
}

impl AbovePlayerBuilder<'_> {
	pub fn make(&self, player: &LocalPlayer, hbar: &HealthBar, time_diff: f32) -> Instance {
		let above_px_pos = self.camera.world_to_pixel_coords(player.get_pos(time_diff) + Vec2::new(-ABOVE_WIDTH / 2.0, 1.25));

		Instance {
			i_pos: [
				2.0 * above_px_pos.x / self.lwinsize.x - 1.0,
				1.0 - 2.0 * above_px_pos.y / self.lwinsize.y,
			],
			i_size: self.size,
			i_t: hbar.get_t(),
			i_colour: hbar.get_colour(),
		}
	}
}
