// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::sync::Arc;

use egui::{Id, SidePanel, Frame, Ui, Margin, FontId, FontFamily, TextStyle, RichText, TextFormat, Color32, Context, text::LayoutJob};

use crate::world::player::colour;
use crate::net::scores::{Scores, ScoreOrdering};

pub(super) const GREY: u8 = 0xdf;

pub fn render(ctx: &Context, scores: &Scores, player_id: u32) {
	let (scores, ordering) = (scores.get_scores(), scores.get_ordering());

	if scores.is_empty() { return; }

	SidePanel::left(Id::new("scores-panel"))
		.frame(Frame::none().inner_margin(Margin::same(16.0)))
		.exact_width(f32::MAX)
		.show_separator_line(false)
		.show(ctx, |ui| {
			ui.label(RichText::new("Scores").color(Color32::from_gray(GREY)).text_style(TextStyle::Name(Arc::from("scores"))));
			ui.add_space(8.0);

			let mut rendered_scores = scores.iter().map(|(&id, &score)| (id, score)).collect::<Vec<_>>();

			// Sorts the scores, if equal the BTreeMap results in them being then sorted by id in ascending order
			match ordering {
				ScoreOrdering::Ascending => rendered_scores.sort_by_key(|e| e.1),
				ScoreOrdering::Descending => rendered_scores.sort_by_key(|e| !e.1), // Not rather than negation to avoid overflow with i32::MIN
			}
			rendered_scores.truncate(16);

			let rs_has_current_player = rendered_scores.iter().any(|e| e.0 == player_id);
			let truncated = rendered_scores.len() < scores.len();

			for (id, score) in rendered_scores {
				add_text(id, score, player_id, ui);
				ui.add_space(4.0);
			}

			if truncated { // If the scores are truncated
				ui.colored_label(Color32::from_gray(GREY), "…");
			}

			// Always shows the current player's score even if not in the top
			if let Some(&score) = scores.get(&player_id) {
				if !rs_has_current_player {
					ui.add_space(4.0);
					add_text(player_id, score, player_id, ui);
				}
			}
		});
}

fn add_text(id: u32, score: i32, player_id: u32, ui: &mut Ui) {
	let mut job = LayoutJob::default();
	job.append(&score.to_string(), if id == player_id { 8.0 } else { 0.0 }, TextFormat { // Indents the score of the current player
		font_id: FontId::new(14.0, FontFamily::Name("scores".into())),
		color: colour::get_gui_colour(id, 0.375),
		.. TextFormat::default()
	});
	ui.label(job);
}
