// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{mem, collections::VecDeque, time::Instant};

use egui::{Context, Window, Frame, Margin, Color32};
use glam::Vec2;
use const_format::formatcp;

use super::health_bar::{HealthBar, HUD_HEIGHT};

use crate::game::config::action::{Action, ActionEvent, ActionState};
use crate::gui::{LABEL_SIZE, layout_job_helper::LayoutJobHelper};
use crate::net::{text_io::{TextIoString, ClientTextInput, ServerTextOutput}, client::Client, message_stream::Security};
use crate::world::player::colour;
use crate::utils::maths::{self, decay::Decay};

#[derive(Default, Clone, Copy)]
enum TextInputState {
	#[default] Disabled,
	Chat,
	Command,
}

impl TextInputState {
	fn open(self) -> bool {
		match self {
			TextInputState::Disabled => false,
			TextInputState::Chat | TextInputState::Command => true,
		}
	}
}

pub struct TextIo { // Chat and commands
	text_input_state: TextInputState,
	current_chat: TextIoString,
	current_command: TextIoString,
	history: VecDeque<(ServerTextOutput, Instant)>,

	pending: Vec<ClientTextInput>,

	display_chat: bool,
	rendered_open: f32,

	prev_rendered_history: usize,
	prev_lwinsize: Vec2,

	ping: u64,
	next_ping: f32,
}

const HISTORY_MAX_LEN: usize = 64;
const HISTORY_MAX_LEN_CLOSED: usize = 6;
const CHAT_PREFIX: char = '>';
const COMMAND_PREFIX: char = '%';

const TEXT_SIZE: f32 = LABEL_SIZE + 3.0;

const LIGHT_GREY: Color32 = Color32::from_gray(0xdf);
const WHITE: Color32 = Color32::WHITE;
const GREEN: Color32 = Color32::from_rgb(0x5f, 0xff, 0x5f);
const LIGHT_GREEN: Color32 = Color32::from_rgb(0xaf, 0xff, 0xaf);
const RED: Color32 = Color32::from_rgb(0xff, 0x5f, 0x5f);
const LIGHT_RED: Color32 = Color32::from_rgb(0xff, 0xaf, 0xaf);
const LIGHT_YELLOW: Color32 = Color32::from_rgb(0xff, 0xff, 0xaf);
const BLUE: Color32 = Color32::from_rgb(0x5f, 0xaf, 0xff);

// Fade out and max times when the text input isn't open
const CLOSED_MAX_TIME: f32 = 60.0;
const CLOSED_FADE_OUT_TIME: f32 = 2.0;

impl TextIo {
	pub fn new() -> TextIo {
		let text_input_state = TextInputState::default();
		TextIo {
			text_input_state,
			current_chat: TextIoString::new(),
			current_command: TextIoString::new(),
			history: VecDeque::with_capacity(HISTORY_MAX_LEN),

			pending: Vec::new(),

			display_chat: true,
			rendered_open: text_input_state.open() as u32 as f32,

			prev_rendered_history: usize::MAX,
			prev_lwinsize: Vec2::NEG_INFINITY,

			ping: 0,
			next_ping: 0.0,
		}
	}

	pub fn action(&mut self, action: ActionEvent) {
		let ActionEvent::Action(action, ActionState::Pressed) = action else { return; };

		self.text_input_state = match (self.text_input_state, action) {
			(TextInputState::Disabled, Action::ChatOpen) => {
				self.display_chat = true;
				TextInputState::Chat
			},
			(TextInputState::Disabled, Action::CommandOpen) => {
				self.display_chat = false;
				TextInputState::Command
			},
			(_, Action::GuiLeave) => TextInputState::Disabled,
			_ => return,
		}
	}

	pub fn open(&self) -> bool {
		self.text_input_state.open()
	}

	pub fn push_history(&mut self, msg: ServerTextOutput) {
		if self.history.len() >= HISTORY_MAX_LEN { self.history.pop_front(); }
		self.history.push_back((msg, Instant::now()));
	}

	pub(super) fn push_current(&mut self, s: &str) {
		let current = match self.text_input_state {
			TextInputState::Chat => &mut self.current_chat,
			TextInputState::Command => &mut self.current_command,
			TextInputState::Disabled => {
				log::warn!("this shouldn't happen");
				return;
			},
		};

		for ch in s.chars() {
			match ch {
				'\r' => {
					let msg = mem::take(current);

					if !msg.trim().is_empty() {
						// Unlikely that an empty chat/command will be needed
						if self.display_chat {
							self.pending.push(ClientTextInput::Chat(msg));
						} else {
							self.pending.push(ClientTextInput::Command(msg));
						}
					}

					self.text_input_state = TextInputState::Disabled;
				},
				'\x08' /* Backspace */ => current.pop(),
				'\x00'..'\x20' | '\x7f' => (), // Prevents any control characters from being pushed (happens when pressing escape or delete)
				_ => current.try_push(ch),
			}
		}
	}

	pub fn pending_messages(&mut self) -> impl Iterator<Item = ClientTextInput> + '_ {
		self.pending.drain(..)
	}

	pub fn render(&mut self, dt: f32, ctx: &Context, lwinsize: Vec2, player_count: usize, client: &Client) {
		const MARGIN: f32 = 8.0;

		// Exponential decay for the text background
		let open = self.open() as u32 as f32;
		self.rendered_open.decay_to(open, 16.0, dt);

		// Prevents never decaying to the exact value from possibly causing rendering problems
		if self.rendered_open < 1e-4 && open == 0.0 {
			self.rendered_open = 0.0;
		} else if self.rendered_open > 1.0 - 1e-4 && open == 1.0 {
			self.rendered_open = 1.0;
		}

		let width = HealthBar::hud_x_px(lwinsize.x).0 - HUD_HEIGHT - MARGIN * 2.0;
		if width <= 0.0 { return; }

		let mut job = LayoutJobHelper::new(TEXT_SIZE).width(width);
		let now = Instant::now();
		let mut rendered_history = 0;

		/*
		 * Hides the prompt of ">" or "%" if the text input is closed and no chat
		 * is visible (actually just uses the largest alpha in the history).
		 */
		let current_line_empty = if self.display_chat { self.current_chat.is_empty() } else { self.current_command.is_empty() };
		let mut current_line_alpha = maths::lerp(if current_line_empty { 0.0 } else { 1.0 }, 1.0, self.rendered_open);

		for (i, (msg, time_received)) in self.history.iter().enumerate() {
			let time = now.duration_since(*time_received).as_secs_f32();
			let alpha_if_closed = if self.history.len() - i > HISTORY_MAX_LEN_CLOSED {
				0.0
			} else if time <= CLOSED_MAX_TIME - CLOSED_FADE_OUT_TIME {
				1.0
			} else if time < CLOSED_MAX_TIME {
				(CLOSED_MAX_TIME - time) / CLOSED_FADE_OUT_TIME
			} else {
				0.0
			};

			let alpha = maths::lerp(alpha_if_closed, 1.0, self.rendered_open);

			if alpha <= 0.0 { continue; }
			current_line_alpha = current_line_alpha.max(alpha);

			job = job.opacity(alpha);
			job = match msg {
				ServerTextOutput::Chat(id, msg) => job.add(&format!("Player {}: ", id + 1), colour::get_gui_colour(*id, 0.4375)).add(msg.as_ref(), LIGHT_GREY),
				ServerTextOutput::Admin(msg) => job.add("Admin: ", RED).add(msg.as_ref(), LIGHT_RED),
				ServerTextOutput::Server(msg) => job.add(msg.as_ref(), WHITE),
				ServerTextOutput::Command(msg) => job.add(formatcp!("{COMMAND_PREFIX} "), GREEN).add(msg.as_ref(), LIGHT_GREEN),
				ServerTextOutput::CommandResponse(msg) => job.add(msg.as_ref(), LIGHT_YELLOW),
			}.newline();

			rendered_history += 1;
		}

		job = job.opacity(current_line_alpha);
		job = if self.display_chat {
			job.add(formatcp!("{CHAT_PREFIX} "), WHITE).add(&self.current_chat, LIGHT_GREY)
		} else {
			job.add(formatcp!("{COMMAND_PREFIX} "), GREEN).add(&self.current_command, LIGHT_GREEN)
		}.newline().newline();

		let spectator_count = client.get_spectator_count().map_or_else(|| String::from("unknown"), |count| count.to_string());

		self.next_ping -= dt;
		if self.next_ping < 0.0 {
			self.ping = client.get_ping().as_millis() as u64;
			self.next_ping += 1.0;
		}

		job = job.opacity(1.0).add(&format!("Players: {player_count}, Spectators: {spectator_count}, Ping: {} ms, ", self.ping), LIGHT_GREY);
		job = match client.get_security() {
			Security::Secure => job.add("Secure connection", GREEN),
			Security::Local => job.add("Local connection", BLUE),
			Security::Insecure => job.add("INSECURE connection! ", RED).add("Chat messages can be eavesdropped.", LIGHT_RED),
		};

		let galley = ctx.fonts(|fonts| fonts.layout_job(job.build()));
		let height = galley.rect.height();

		let alpha = (0x9f as f32 * self.rendered_open).round() as u8;

		// Prevents first-frame jitter when the history length changes or window size changes
		if rendered_history != self.prev_rendered_history || lwinsize != self.prev_lwinsize {
			ctx.request_discard("");
			self.prev_rendered_history = rendered_history;
			self.prev_lwinsize = lwinsize;
		}

		Window::new("text-panel")
			.frame(Frame::none().inner_margin(Margin::same(MARGIN)).fill(Color32::from_black_alpha(alpha)))
			.fixed_pos([0.0, lwinsize.y - (height + MARGIN * 2.0)])
			.title_bar(false)
			.show(ctx, |ui| {
				ui.set_width(width);
				ui.set_height(height);
				ui.label(galley);
			});
	}
}
