// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::num::NonZeroUsize;
use std::fmt::{Debug, Formatter, Result as FmtResult};
use std::collections::VecDeque;

pub struct RollingAverage {
	capacity: usize,
	queue: VecDeque<f32>,
	sum: f32, // Redundant information cached for optimisation
}

impl RollingAverage {
	pub fn new(capacity: NonZeroUsize) -> RollingAverage {
		let capacity = capacity.into();
		RollingAverage {
			capacity,
			queue: VecDeque::with_capacity(capacity),
			sum: 0.0,
		}
	}

	#[cfg(feature = "client")]
	pub fn reset(&mut self) {
		self.queue.clear();
		self.sum = 0.0;
	}

	pub fn add(&mut self, x: f32) {
		if self.queue.len() == self.capacity {
			self.sum -= self.queue.pop_back().unwrap();
		}
		self.sum += x;
		self.queue.push_front(x);
	}

	#[cfg(feature = "client")]
	pub fn mean(&self) -> Option<f32> {
		if self.queue.is_empty() { None } else { Some(self.sum / self.queue.len() as f32) }
	}

	pub fn mean_and_stddev(&self) -> Option<(f32, f32)> {
		if self.queue.is_empty() { return None; }

		let mean = self.sum / self.queue.len() as f32;
		let stddev = (self.queue.iter().map(|&x| (x - mean).powi(2)).sum::<f32>() / self.queue.len() as f32).sqrt();
		Some((mean, stddev))
	}
}

impl Debug for RollingAverage {
	fn fmt(&self, fmt: &mut Formatter<'_>) -> FmtResult {
		self.queue.fmt(fmt)
	}
}
