// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
#[cfg(feature = "client")] use glam::Vec2;

use super::{DMap, DSet, Player, bullet::Bullet, powerup::{PowerupType, Effect}, event::move_chooser::MoveType};
#[cfg(feature = "client")] use super::player::direction::Direction;

#[cfg(feature = "client")] use crate::utils::maths::Circle;

/*
 * Both the client and server need to know certain things that have changed
 * after updating the world. However, the exact information that's relevant to
 * them differs.
 *
 * As an optimisation, generics are used to avoid performing unnecessary
 * operations when calling `update`.
 */
pub trait Changes {
	/*
	 * For the client, prevents changes reported to old players from being
	 * visualised on players (in case if a player rapidly exits and joins).
	 *
	 * For the server, probably not needed for now but I'm adding it for good
	 * practice to clean things up when players are removed.
	 */
	fn removed_players(&mut self, old_ids: DSet<u32>);

	fn add_hit(&mut self, bullet: Bullet, player: &Player, player_id: u32, killed: bool);
	fn pos_changed(&mut self, id: u32);
	fn ammo_crate_collected(&mut self, id: u32);
	fn powerup_collected(&mut self, id: u32, typ: PowerupType);
	fn player_reset(&mut self, id: u32);
	fn items_reset(&mut self);
}

#[cfg(feature = "client")]
pub struct Hit {
	pub bullet: Bullet,
	pub player_id: u32,
	pub player_pos: Vec2,
	pub player_vel: Vec2,
	pub player_dir: Direction,
	pub killed: bool,
}

#[cfg(feature = "client")]
#[derive(Default)]
pub struct ClientChanges { // Used by the clients
	pub moved: DSet<u32>, // Whenever the server moves a player, which can be on death or teleportation
	pub hits: Vec<Hit>,
	pub reset: DSet<u32>,
	pub items_reset: bool,
}

#[cfg(feature = "client")]
impl Changes for ClientChanges {
	fn removed_players(&mut self, old_ids: DSet<u32>) {
		for id in &old_ids {
			self.moved.remove(id);
			self.reset.remove(id);
		}

		self.hits.retain(|hit| !old_ids.contains(&hit.player_id));
	}

	fn add_hit(&mut self, bullet: Bullet, player: &Player, player_id: u32, killed: bool) {
		self.hits.push(Hit {
			bullet,
			player_id,
			player_pos: player.get_pos(),
			player_vel: player.get_vel(),
			player_dir: player.get_dir(),
			killed,
		});
	}

	fn pos_changed(&mut self, id: u32) { self.moved.insert(id); }
	fn ammo_crate_collected(&mut self, _id: u32) {}
	fn powerup_collected(&mut self, _id: u32, _typ: PowerupType) {}
	fn player_reset(&mut self, id: u32) { self.reset.insert(id); }
	fn items_reset(&mut self) { self.items_reset = true; }
}

#[derive(Default)]
pub struct ServerChanges {
	pub kills: Vec<Kill>, // Player ids duplicated if they get multiple kills in a single update
	pub(super) randomly_moved: DMap<u32, MoveType>,
	pub(super) dead: DSet<u32>,
	pub(super) ammo_crates_collected: Vec<u32>,
	pub(super) effects_to_apply: Vec<(u32, Effect)>,
}

pub struct Kill {
	pub killer: Option<u32>,
	pub killed: u32,
}

impl Changes for ServerChanges {
	fn removed_players(&mut self, old_ids: DSet<u32>) {
		for id in &old_ids {
			self.randomly_moved.remove(id);
			self.dead.remove(id);
		}

		self.ammo_crates_collected.retain(|id| !old_ids.contains(id));
		self.effects_to_apply.retain(|(id, _)| !old_ids.contains(id));
		// Not changing self.kills as that is used before the players are eliminated
	}

	fn add_hit(&mut self, bullet: Bullet, _player: &Player, player_id: u32, killed: bool) {
		if killed {
			self.kills.push(Kill {
				killer: bullet.get_owner(),
				killed: player_id,
			});

			self.randomly_moved.insert(player_id, MoveType::Spawn);
			self.dead.insert(player_id);
		}
	}

	fn pos_changed(&mut self, _id: u32) {}
	fn ammo_crate_collected(&mut self, id: u32) { self.ammo_crates_collected.push(id); }

	fn powerup_collected(&mut self, id: u32, typ: PowerupType) {
		if let Some(effect) = Effect::try_from_type(typ) {
			self.effects_to_apply.push((id, effect));
		} else if typ == PowerupType::Teleportation {
			self.randomly_moved.insert(id, MoveType::Teleportation);
		}
	}

	fn player_reset(&mut self, _id: u32) {}
	fn items_reset(&mut self) {}
}

pub struct IgnoreChanges;

impl Changes for IgnoreChanges {
	fn removed_players(&mut self, _old_ids: DSet<u32>) {}
	fn add_hit(&mut self, _bullet: Bullet, _player: &Player, _player_id: u32, _killed: bool) {}
	fn pos_changed(&mut self, _id: u32) {}
	fn ammo_crate_collected(&mut self, _id: u32) {}
	fn powerup_collected(&mut self, _id: u32, _typ: PowerupType) {}
	fn player_reset(&mut self, _id: u32) {}
	fn items_reset(&mut self) {}
}
