// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
pub mod move_chooser;
pub mod portal;
pub mod effect_zone;
pub mod effect_controller;

use serde::{Serialize, Deserialize};
use glam::Vec2;
use rand::{RngCore, SeedableRng};

use move_chooser::{MoveChooser, MoveType, MoveHint};
use effect_controller::{EffectController, underpower::UnderpowerCommand};
use portal::PortalManager;
use effect_zone::EffectZones;

use super::{
	World, WorldRng, changes::{Changes, ServerChanges, IgnoreChanges},
	player::{Player, ammo_count::AmmoCount}, ammo_crate::{AmmoCrate, RADIUS as AMMO_CRATE_RADIUS},
	powerup::{Powerup, PowerupType, Effect, RADIUS as POWERUP_RADIUS, active_effects::EffectInfo},
	DELTA_TIME, MAX_AMMO_CRATES, MAX_POWERUPS
};

use crate::blocks::Blocks;
use crate::net::{server::{config::EventConfig, joined::Joined}, text_io::TextIoString};
use crate::utils::{rng::Float, maths::{self, Circle, CollidingRect, RectCorners}};

#[derive(Clone, Serialize, Deserialize)]
pub enum WorldEvent {
	/// When a new player joins.
	NewPlayer(u32),

	/// When a player is removed. Most of the time when a player is removed, this
	/// event isn't emitted but rather the player id isn't included in the update
	/// map.
	///
	/// This event is useful when removing a player after an update has completed
	/// but that player should exist during the update.
	RemovePlayer(u32),

	/// Calls `Player::reset`, which reset's the health and attack cooldown.
	PlayerReset(u32),

	/// Sets the player's position.
	PlayerPos(u32, Vec2),

	/// Sets the player's velocity.
	PlayerVel(u32, Vec2),

	/// Sets the player's amount of ammo.
	PlayerAmmo(u32, AmmoCount),

	/// Sets the player's health.
	PlayerHealth(u32, f32),

	/// Adds an effect to a player.
	PlayerEffect {
		id: u32,
		effect: Effect,
		info: EffectInfo,
	},

	/// Adds a new ammo crate. The amount of ammo is determined when the player
	/// picks it up, in which the server may give the player ammo with the
	/// `PlayerAmmo` update.
	NewAmmoCrate(Vec2),

	/// Adds a new powerup.
	NewPowerup(Vec2, PowerupType),

	/// Removes all bullets, ammo crates and powerups from the world
	ClearItems,
}

pub struct EventChooser {
	rng: WorldRng,
	effect_controller: EffectController,
	portal_manager: PortalManager,
	effect_zones: EffectZones,
	config: EventConfig,
	ppssr2: f32, // Player powerup spawn separation radius squared
}

impl EventChooser {
	pub fn new(portal_manager: PortalManager, effect_zones: EffectZones, config: EventConfig) -> EventChooser {
		let rng = WorldRng::seed_from_u64(rand::random::<u64>());
		let ppssr2 = config.player_powerup_spawn_separation_radius.powi(2);
		EventChooser { rng, effect_controller: EffectController::new(config.powerup_effect_reduction), portal_manager, effect_zones, config, ppssr2 }
	}

	pub fn underpower_command(&mut self, id: u32, cmd: UnderpowerCommand) -> TextIoString {
		self.effect_controller.underpower_command(id, cmd)
	}

	pub fn get_pre_events(&mut self, world: &World, blocks: &Blocks, move_chooser: &dyn MoveChooser, joined: Option<&Joined>) -> Vec<WorldEvent> {
		/*
		 * Also using the client limits to prevent clients from disconnecting if
		 * there are too many. Not also doing this with bullets as that case is
		 * less likely to occur (this can occur by players staying idle for a
		 * while) and players not being able to shoot would be weird.
		 */
		let mut events = Vec::new();

		if self.rng.next_f32() < EventChooser::get_prob(self.config.ammo_crate_period, world.players.len()) && world.ammo_crates.len() < MAX_AMMO_CRATES {
			let can_spawn = |pos| {
				let rect = RectCorners(pos - AMMO_CRATE_RADIUS, pos + AMMO_CRATE_RADIUS);
				!(blocks.iter_range(pos, AMMO_CRATE_RADIUS).count() > 0 || self.portal_manager.sink_colliding(&rect))
			};
			if let Some(pos) = EventChooser::find_item_pos(world, &mut self.rng, &self.config, AMMO_CRATE_RADIUS, can_spawn) {
				events.push(WorldEvent::NewAmmoCrate(pos));
			}
		}

		if self.rng.next_f32() < EventChooser::get_prob(self.config.powerup_period, world.players.len()) && world.powerups.len() < MAX_POWERUPS && !self.config.allowed_powerups.is_empty() {
			let can_spawn = |pos| {
				let circle = (pos, POWERUP_RADIUS);
				!(
					blocks.iter_range(pos, POWERUP_RADIUS).any(|block| circle.colliding_rect(block.get_rect())) ||
					self.portal_manager.sink_colliding(&circle) ||
					world.players.values().any(|player| pos.distance_squared(player.get_pos()) < self.ppssr2)
				)
			};
			if let Some(pos) = EventChooser::find_item_pos(world, &mut self.rng, &self.config, POWERUP_RADIUS, can_spawn) {
				let effect = self.config.allowed_powerups[self.rng.next_u32() as usize % self.config.allowed_powerups.len()];
				events.push(WorldEvent::NewPowerup(pos, effect));
			}
		}

		if let Some(joined) = joined {
			for id in joined.players() {
				self.add_player(id, world, blocks, move_chooser, &mut events);
			}
		}

		events
	}

	pub fn add_player(&self, id: u32, world: &World, blocks: &Blocks, move_chooser: &dyn MoveChooser, events: &mut Vec<WorldEvent>) {
		events.push(WorldEvent::NewPlayer(id));
		events.push(WorldEvent::PlayerPos(id, self.find_player_pos(MoveType::Spawn, id, world, blocks, move_chooser)));
		events.push(WorldEvent::PlayerAmmo(id, self.config.ammo_count.0));
	}

	fn apply_and_push(world: &mut World, events: &mut Vec<WorldEvent>, event: WorldEvent) {
		world.apply_event(&event, &mut IgnoreChanges);
		events.push(event);
	}

	fn update_effect_controller(&mut self, world: &mut World, events: &mut Vec<WorldEvent>) {
		let mut tmp_events = Vec::new();
		self.effect_controller.update(world, &mut tmp_events, DELTA_TIME);
		for event in tmp_events {
			Self::apply_and_push(world, events, event);
		}
	}

	pub fn get_and_apply_post_events(&mut self, world: &mut World, blocks: &Blocks, move_chooser: &dyn MoveChooser, changes: &ServerChanges) -> Vec<WorldEvent> {
		let mut events = Vec::new();

		for (&id, &typ) in &changes.randomly_moved {
			Self::apply_and_push(world, &mut events, WorldEvent::PlayerPos(id, self.find_player_pos(typ, id, world, blocks, move_chooser)));
			self.portal_manager.on_random_move(id);
		}

		for &id in &changes.dead {
			Self::apply_and_push(world, &mut events, WorldEvent::PlayerVel(id, Vec2::ZERO));
			Self::apply_and_push(world, &mut events, WorldEvent::PlayerAmmo(id, self.config.ammo_count.0));

			// Resets all effects when a player dies
			self.effect_controller.reset_player_effects(id);
		}

		{
			let mut tmp_events = Vec::new();
			self.portal_manager.events(world, blocks, &mut self.rng, &mut tmp_events);
			for event in tmp_events { Self::apply_and_push(world, &mut events, event); }
		}

		for &id in &changes.ammo_crates_collected {
			if changes.dead.contains(&id) { continue; }
			let Some(player) = world.players.get(&id) else { continue; };

			let range = self.config.ammo_crate_supply_range;
			let (min, max) = (range.0, range.1);
			let new_ammo = player.get_ammo_count() + (min + self.rng.next_u32() % ((max - min).saturating_add(1)));
			Self::apply_and_push(world, &mut events, WorldEvent::PlayerAmmo(id, new_ammo));
		}

		for &(id, effect) in &changes.effects_to_apply {
			if changes.dead.contains(&id) { continue; }
			let Some(player) = world.players.get(&id) else { continue; };

			self.effect_controller.powerup_collected(id, effect, EffectInfo { time: self.config.effect_time, power: self.config.get_effect_power(effect) });

			if effect == Effect::Regeneration {
				Self::apply_and_push(world, &mut events, WorldEvent::PlayerHealth(id, player.add_health(self.config.health_powerup_health_increase)));
			}
		}

		self.effect_zones.update(world, &mut self.effect_controller);
		self.update_effect_controller(world, &mut events);
		events
	}

	pub fn reset_get_and_apply_events(&mut self, world: &mut World, blocks: &Blocks, move_chooser: &dyn MoveChooser) -> Vec<WorldEvent> {
		let mut events = Vec::with_capacity(1 + world.players.len() * 4);

		Self::apply_and_push(world, &mut events, WorldEvent::ClearItems);
		#[allow(clippy::needless_collect)] // It actually is necessary, you get a compiler error if you remove it
		for id in world.players.keys().copied().collect::<Vec<_>>() {
			Self::apply_and_push(world, &mut events, WorldEvent::PlayerReset(id));

			let mut tmp_events = Vec::new();
			self.add_player(id, world, blocks, move_chooser, &mut tmp_events);
			for event in tmp_events {
				Self::apply_and_push(world, &mut events, event);
			}
		}

		self.effect_controller.reset_all_effects();
		self.update_effect_controller(world, &mut events);

		self.portal_manager.reset();
		events
	}

	fn get_prob(mean_interval: f32, players: usize) -> f32 {
		(DELTA_TIME / mean_interval) * players as f32
	}

	fn find_item_pos(world: &World, rng: &mut WorldRng, config: &EventConfig, radius: f32, should_spawn: impl Fn(Vec2) -> bool) -> Option<Vec2> {
		debug_assert!(!world.players.is_empty()); // Nothing should spawn when there are no players

		for _ in 0..20 { // Gives up if no suitable place is found after too many tries
			// Lazy O(n) algorithm, might improve later
			let rand_player_pos = world.players.iter().nth(rng.next_u32() as usize % world.players.len()).unwrap().1.get_pos();

			// Gets the range for spawning, which must be inside the world
			let min = (rand_player_pos - Vec2::splat(config.item_spawn_radius)).max(radius - world.size / 2.0);
			let max = (rand_player_pos + Vec2::splat(config.item_spawn_radius)).min(world.size / 2.0 - radius);

			let pos = Vec2::new(maths::lerp(min.x, max.x, rng.next_f32()), maths::lerp(min.y, max.y, rng.next_f32()));

			// If not colliding with any block (or some other condition)
			if should_spawn(pos) {
				return Some(pos);
			}
		}

		None
	}

	/**
	 * Makes an attempt to not spawn the player on a powerup, but if it no
	 * position is found after too many attempts, it is used anyway.
	 */
	fn find_player_pos(&self, typ: MoveType, id: u32, world: &World, blocks: &Blocks, move_chooser: &dyn MoveChooser) -> Vec2 {
		let mut i = 0;
		loop {
			let (pos, hint) = move_chooser.generate(typ, id, blocks);

			if i >= 20 || hint == Some(MoveHint::Constant) /* Small optimisation if it's known to stay the same */ { return pos; } // Prevents an infinite loop

			if world.powerups.iter().any(|powerup| pos.distance_squared(powerup.get_pos()) < self.ppssr2) { i += 1; }
			else { return pos; }
		}
	}
}

impl World {
	fn apply_event(&mut self, event: &WorldEvent, changes: &mut impl Changes) {
		match *event {
			WorldEvent::NewPlayer(id) => _ = self.players.insert(id, Player::new()),
			WorldEvent::RemovePlayer(id) => _ = self.players.remove(&id),
			WorldEvent::PlayerReset(id) => {
				self.player_present(id, Player::reset);
				changes.player_reset(id);
			},
			WorldEvent::PlayerPos(id, pos) => {
				self.player_present(id, |player| player.set_pos(pos));
				changes.pos_changed(id);
			},
			WorldEvent::PlayerVel(id, vel) => self.player_present(id, |player| player.set_vel(vel)),
			WorldEvent::PlayerHealth(id, health) => self.player_present(id, |player| player.set_health(health)),
			WorldEvent::PlayerAmmo(id, ammo) => self.player_present(id, |player| player.set_ammo_count(ammo)),
			WorldEvent::PlayerEffect { id, effect, info } => self.effects.add_effect(id, effect, info),
			WorldEvent::NewAmmoCrate(pos) => self.ammo_crates.push(AmmoCrate::new(pos)),
			WorldEvent::NewPowerup(pos, typ) => self.powerups.push(Powerup::new(pos, typ)),
			WorldEvent::ClearItems => {
				self.bullets.clear();
				self.ammo_crates.clear();
				self.powerups.clear();
				changes.items_reset();
			},
		}
	}

	pub fn apply_events(&mut self, events: &[WorldEvent], changes: &mut impl Changes) {
		for event in events {
			self.apply_event(event, changes);
		}
	}

	fn player_present(&mut self, id: u32, f: impl Fn(&mut Player)) {
		if let Some(player) = self.players.get_mut(&id) {
			f(player);
		}
	}
}
