// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::rc::Rc;

use glium::{Display, Frame, DrawParameters, Texture2d};
use glutin::surface::WindowSurface;
use glam::Mat4;

use super::{World, SmoothWorld, border_renderer::BorderRenderer, player::{PlayerRenderer, forcefield::ForcefieldRenderer}, bullet::BulletRenderer, ammo_crate::AmmoCrateRenderer, powerup::PowerupRenderer, special_area::SpecialAreaRenderer};

use crate::game::filesystem::Filesystem;
use crate::playing::{model::Model, view::Camera};
use crate::blocks::{Blocks, BlocksRenderer};
use crate::utils::blending::{ALPHA_BLENDING, ADDITIVE_BLENDING};

const MIN_TINT: f32 = 0.4375; // Brightness of the initial tint
const TINT_TIME: f32 = 0.15625;

pub struct Renderer {
	border: BorderRenderer,
	players: PlayerRenderer,
	bullets: BulletRenderer,
	ammo_crates: AmmoCrateRenderer,
	powerups: PowerupRenderer,
	forcefields: ForcefieldRenderer,
	blocks: BlocksRenderer,
	special_areas: SpecialAreaRenderer,
}

impl Renderer {
	pub fn new(display: &Display<WindowSurface>, fs: &Filesystem, player_texture: Rc<Texture2d>) -> Renderer {
		Renderer {
			border: BorderRenderer::new(display, fs),
			players: PlayerRenderer::new(display, fs, player_texture),
			bullets: BulletRenderer::new(display, fs),
			ammo_crates: AmmoCrateRenderer::new(display, fs),
			powerups: PowerupRenderer::new(display, fs),
			forcefields: ForcefieldRenderer::new(display, fs),
			blocks: BlocksRenderer::new(display, fs),
			special_areas: SpecialAreaRenderer::new(display, fs),
		}
	}

	pub fn reset(&mut self, display: &Display<WindowSurface>, model: &Model) {
		self.border.reset(model.get_world());
		self.blocks.reset(model.get_blocks());
		self.special_areas.update(&model.get_world().special_areas, display);
	}

	pub fn render_border(&self, frame: &mut Frame, params: &DrawParameters, matrix: &Mat4) {
		self.border.render(frame, params, matrix);
	}

	pub fn render_special_areas(&self, frame: &mut Frame, params: &DrawParameters, matrix: &Mat4) {
		self.special_areas.render(frame, params, matrix);
	}

	#[allow(clippy::too_many_arguments)]
	pub fn render(&mut self, dt: f32,
		display: &Display<WindowSurface>, frame: &mut Frame, params: &mut DrawParameters,
		time_diff: f32, world: &World, smooth_world: &mut SmoothWorld, blocks: &Blocks, camera: &Camera,
	) {
		let matrix = camera.get_matrix();
		let time_shift = smooth_world.get_total_time_shift() + time_diff;

		self.forcefields.render(&smooth_world.get_forcefields(world, time_diff), display, frame, params, matrix);
		self.ammo_crates.render(world.get_ammo_crates(), display, frame, params, matrix, time_shift);
		params.blend.color = ADDITIVE_BLENDING; // The powerups are basically "lightning balls" so additive blending can be used (also makes texture generation easier)
		self.powerups.render(world.get_powerups(), display, frame, params, matrix, time_shift);
		params.blend.color = ALPHA_BLENDING;

		self.bullets.render(smooth_world.get_bullets(), display, frame, params, matrix, time_diff);

		// Renders the user's player above all other players
		for (id, player) in smooth_world.player_iter_mut() {
			let tint = if player.last_hurt < TINT_TIME {
				(player.last_hurt / TINT_TIME) * (1.0 - MIN_TINT) + MIN_TINT
			} else {
				1.0
			};

			player.hbar.update(&player.player, dt);
			self.players.push(player, id, tint, time_diff);
			player.last_hurt += dt; // Want to update the tint after rendering so the darkest tint is rendered once
		}
		self.players.render(display, frame, params, matrix);

		// Renders the blocks over everything else
		self.blocks.render(display, frame, params, blocks, camera);
	}
}
